"""Operators for common transformations including different types of waves."""

import numpy as np

import treams.special as sc
from treams import util
from treams._lattice import Lattice, WaveVector
from acoustotreams._materialacoustics import AcousticMaterial
import acoustotreams._coreacoustics as core
from treams._operators import Operator,FieldOperator
import acoustotreams._wavesacoustics as wv
import acoustotreams.ssw as ssw
import acoustotreams.spw as spw
import acoustotreams.scw as scw


def _ssw_rotate(phi, theta, psi, basis, to_basis, where):
    """Rotate scalar spherical waves."""
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    res = ssw.rotate(
        *(m[:, None] for m in to_basis.lm), *basis.lm, phi, theta, psi, where=where
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(res, basis=(to_basis, basis))

def _scw_rotate(phi, basis, to_basis, where):
    """Rotate scalar cylindrical waves."""
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    res = scw.rotate(*(m[:, None] for m in to_basis.zm), *basis.zm, phi, where=where)
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(res, basis=(to_basis, basis))


def _spwa_rotate(phi, basis, where):
    """Rotate scalar plane waves (actually rotates the basis)."""
    c1, s1 = np.cos(phi), np.sin(phi)
    r = np.array([[c1, -s1, 0], [s1, c1, 0], [0, 0, 1]])
    kvecs = r @ np.array([basis.qx, basis.qy, basis.qz])
    res = np.eye(len(basis))
    res[..., np.logical_not(where)] = 0
    newbasis = core.ScalarPlaneWaveBasisByUnitVector(zip(*kvecs))
    if basis.lattice is not None:
        newbasis.lattice = basis.lattice.rotate(phi)
    if basis.kpar is not None:
        newbasis.kpar = basis.kpar.rotate(phi)
    return core.AcousticsArray(res, basis=(newbasis, basis))


def _spwp_rotate(phi, basis, where):
    """Rotate partial scalar plane waves (actually rotates the basis)."""
    if basis.alignment != "xy":
        raise ValueError(f"rotation on alignment: '{basis.alignment}'")
    c1, s1 = np.cos(phi), np.sin(phi)
    r = np.array([[c1, -s1], [s1, c1]])
    kx, ky = basis[()]
    res = np.eye(len(basis))
    res[..., np.logical_not(where)] = 0
    newbasis = core.ScalarPlaneWaveBasisByComp(zip(*(r @ np.array([kx, ky]))))
    if basis.lattice is not None:
        newbasis.lattice = basis.lattice.rotate(phi)
    if basis.kpar is not None:
        newbasis.kpar = basis.kpar.rotate(phi)
    return core.AcousticsArray(res, basis=(newbasis, basis))


def rotate(phi, theta=0, psi=0, *, basis, where=True):
    """Rotation matrix.

    For the given Euler angles apply a rotation for the given basis. In some basis sets
    only rotations around the z-axis are permitted.

    Args:
        phi (float): First Euler angle (rotation about z)
        theta (float, optional): Second Euler angle (rotation about y)
        psi (float, optional): Third Euler angle (rotation about z)
        basis (:class:`ScalarBasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        where (array-like, bool, optional): Only evaluate parts of the rotation matrix,
            the given array must have a shape that matches the output shape.
    """
    if isinstance(basis, (tuple, list)):
        to_basis, basis = basis
    else:
        to_basis = basis

    if isinstance(basis, core.ScalarSphericalWaveBasis):
        return _ssw_rotate(phi, theta, psi, basis, to_basis, where)
    if theta != 0:
        raise ValueError("non-zero theta for rotation")
    phi = phi + psi
    if isinstance(basis, core.ScalarCylindricalWaveBasis):
        return _scw_rotate(phi, basis, to_basis, where)
    if to_basis != basis:
        raise ValueError("invalid basis")
    if isinstance(basis, core.ScalarPlaneWaveBasisByComp):
        return _spwp_rotate(phi, basis, where)
    if isinstance(basis, core.ScalarPlaneWaveBasisByUnitVector):
        return _spwa_rotate(phi, basis, where)
    raise TypeError("invalid basis")


class Rotate(Operator):
    _FUNC = staticmethod(rotate)

    def __init__(self, phi, theta=0, psi=0, *, isinv=False):
        super().__init__(phi, theta, psi, isinv=isinv)

    def _call_inv(self, **kwargs):
        if "basis" in kwargs and isinstance(kwargs["basis"], tuple):
            kwargs["basis"] = kwargs["basis"][::-1]
        return self.FUNC(*(-a for a in self._args[::-1]), **kwargs)
    

def _ssw_translate(r, basis, to_basis, k0, material, where):
    """Translate scalar spherical waves."""
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    ks = k0 * material.nmp
    r = sc.car2sph(r)
    res = ssw.translate(
        *(m[:, None] for m in to_basis.lm),
        *basis.lm,
        ks * r[..., None, None, 0],
        r[..., None, None, 1],
        r[..., None, None, 2],
        singular=False,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(
        res,
        k0=(k0, k0),
        basis=(to_basis, basis),
        material=(material, material),
    )

def _scw_translate(r, basis, k0, to_basis, material, where):
    """Translate scalar cylindrical waves."""
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    ks = material.ks(k0)
    krhos = np.sqrt(ks * ks - basis.kz * basis.kz + 0j)
    krhos[krhos.imag < 0] = -krhos[krhos.imag < 0]
    r = sc.car2cyl(r)
    res = scw.translate(
        *(m[:, None] for m in to_basis.zm),
        *basis.zm,
        krhos * r[..., None, None, 0],
        r[..., None, None, 1],
        r[..., None, None, 2],
        singular=False,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(
        res, k0=(k0, k0), basis=(to_basis, basis), material=(material, material)
    )

def _spw_translate(r, basis, k0, to_basis, material, modetype, where):
    """Translate scalar plane waves."""
    kvecs = basis.kvecs(k0, material, modetype)
    kx, ky, kz = to_basis.kvecs(k0, material, modetype)
    where = (
        where
        & (np.abs(kx[:, None] - kvecs[0]) < 1e-14)
        & (np.abs(ky[:, None] - kvecs[1]) < 1e-14)
        & (np.abs(kz[:, None] - kvecs[2]) < 1e-14)
    )
    res = spw.translate(
        *kvecs,
        r[..., None, None, 0],
        r[..., None, None, 1],
        r[..., None, None, 2],
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(
        res,
        k0=(k0,) * 2,
        basis=(to_basis, basis),
        material=(material,) * 2,
        modetype=(modetype,) * 2,
    )

def translate(
    r, *, basis, k0=None, material=AcousticMaterial(), modetype=None, where=True
):
    """Translation matrix.

    Translate the given basis modes along the translation vector.

    Args:
        r (array-like): Translation vector
        basis (:class:`ScalarBasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        k0 (float, optional): Wave number.
        material (:class:`~acoustotreams.AcousticMaterial` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode, only used for
            :class:`ScalarPlaneWaveBasisByComp`.
        where (array-like, bool, optional): Only evaluate parts of the translation
            matrix, the given array must have a shape that matches the output shape.
    """
    if isinstance(basis, (tuple, list)):
        to_basis, basis = basis
    else:
        to_basis = basis
    material = AcousticMaterial(material)

    r = np.asanyarray(r)
    if r.shape[-1] != 3:
        raise ValueError("invalid 'r'")
    if isinstance(basis, core.ScalarPlaneWaveBasis):
        if isinstance(basis, core.ScalarPlaneWaveBasisByComp):
            modetype = "up" if modetype is None else modetype
        return _spw_translate(r, basis, k0, to_basis, material, modetype, where)
    if isinstance(basis, core.ScalarSphericalWaveBasis):
        return _ssw_translate(r, basis, to_basis, k0, material, where)
    if isinstance(basis, core.ScalarCylindricalWaveBasis):
        return _scw_translate(r, basis, k0, to_basis, material, where)
    raise TypeError("invalid basis")


class Translate(Operator):
    """Translation matrix.

    When called as attribute of an object it returns a suitable translation matrix to
    transform it. See also :func:`translate`.
    """

    _FUNC = staticmethod(translate)

    def __init__(self, r, *, isinv=False):
        super().__init__(r, isinv=isinv)

    def _call_inv(self, **kwargs):
        if "basis" in kwargs and isinstance(kwargs["basis"], tuple):
            kwargs["basis"] = kwargs["basis"][::-1]
        return self.FUNC(np.negative(self._args[0]), **kwargs)


def _ssw_ssw_expand(basis, to_basis, to_modetype, k0, material, modetype, where):
    """Expand scalar spherical waves into scalar spherical waves."""
    if not (
        modetype == "regular" == to_modetype
        or modetype == "singular" == to_modetype
        or (modetype == "singular" and to_modetype == "regular")
    ):
        raise ValueError(f"invalid expansion from {modetype} to {to_modetype}")
    rs = sc.car2sph(to_basis.positions[:, None, :] - basis.positions)
    ks = k0 * AcousticMaterial().c / material.c
    res = ssw.translate(
        *(m[:, None] for m in to_basis.lm),
        *basis.lm,
        ks * rs[to_basis.pidx[:, None], basis.pidx, 0],
        rs[to_basis.pidx[:, None], basis.pidx, 1],
        rs[to_basis.pidx[:, None], basis.pidx, 2],
        singular=modetype != to_modetype,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    res = core.AcousticsArray(
        res, k0=k0, basis=(to_basis, basis), material=material
    )
    if modetype == "singular" and to_modetype == "regular":
        res.modetype = (to_modetype, modetype)
    return res

def _ssw_scw_expand(basis, to_basis, k0, material, where):
    """Expand scalar cylindrical waves into scalar spherical waves."""
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    ks = material.ks(k0)
    res = scw.to_ssw(
        *(m[:, None] for m in to_basis.lm),
        *basis.zm,
        ks,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(
        res,
        k0=k0,
        basis=(to_basis, basis),
        material=material,
        modetype=("regular", "regular"),
    )

def _ssw_spw_expand(basis, to_basis, k0, material, modetype, where):
    """Expand scalar plane waves into scalar spherical waves."""
    if isinstance(basis, core.ScalarPlaneWaveBasisByComp):
        modetype = "up" if modetype is None else modetype
    kvecs = basis.kvecs(k0, material, modetype)
    res = spw.to_ssw(
        *(m[:, None] for m in to_basis.lm),
        *kvecs,
        where=where,
    ) * spw.translate(
        *kvecs,
        to_basis.positions[to_basis.pidx, None, 0],
        to_basis.positions[to_basis.pidx, None, 1],
        to_basis.positions[to_basis.pidx, None, 2],
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(
        res,
        basis=(to_basis, basis),
        k0=k0,
        material=material,
        modetype=("regular", modetype),
    )

def _scw_scw_expand(basis, to_basis, to_modetype, k0, material, modetype, where):
    """Expand scalar cylindrical waves into scalar cylindrical waves."""
    rs = sc.car2cyl(to_basis.positions[:, None, :] - basis.positions)
    krhos = material.krhos(k0, basis.kz)
    res = scw.translate(
        *(m[:, None] for m in to_basis.zm),
        *basis.zm,
        krhos * rs[to_basis.pidx[:, None], basis.pidx, 0],
        rs[to_basis.pidx[:, None], basis.pidx, 1],
        rs[to_basis.pidx[:, None], basis.pidx, 2],
        singular=modetype != to_modetype,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    res = core.AcousticsArray(res, k0=k0, basis=(to_basis, basis), material=material)
    if modetype == "singular" and to_modetype == "regular":
        res.modetype = (to_modetype, modetype)
    return res

def _scw_spw_expand(basis, to_basis, k0, material, modetype, where):
    """Expand scalar plane waves into scalar cylindrical waves."""
    if isinstance(basis, core.ScalarPlaneWaveBasisByComp):
        modetype = "up" if modetype is None else modetype
    kvecs = basis.kvecs(k0, material, modetype)
    res = spw.to_scw(
        *(m[:, None] for m in to_basis.zm), *kvecs, where=where
    ) * spw.translate(
        *kvecs,
        to_basis.positions[to_basis.pidx, None, 0],
        to_basis.positions[to_basis.pidx, None, 1],
        to_basis.positions[to_basis.pidx, None, 2],
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(
        res,
        basis=(to_basis, basis),
        k0=k0,
        material=material,
        modetype=("regular", modetype),
    )


def _spw_spw_expand(basis, to_basis, k0, material, modetype, where):
    """Expand scalar plane waves into scalar plane waves."""
    if isinstance(basis, core.ScalarPlaneWaveBasisByComp):
        modetype = "up" if modetype is None else modetype
    kvecs = basis.kvecs(k0, material, modetype)
    if isinstance(to_basis, core.ScalarPlaneWaveBasisByComp):
        modetype = "up" if modetype is None else modetype
    kx, ky, kz = to_basis.kvecs(k0, material, modetype)
    res = np.array(
        where
        & (kx[:, None] == kvecs[0])
        & (ky[:, None] == kvecs[1])
        & (kz[:, None] == kvecs[2]),
        int,
    )
    return core.AcousticsArray(
        res, basis=(to_basis, basis), k0=k0, material=material, modetype=modetype
    )


def expand(
    basis, modetype=None, *, k0=None, material=AcousticMaterial(), where=True
):
    """Expansion matrix.

    Expand the modes from one basis set to another basis set. If applicable the modetype
    can also be changed, like for spherical and cylindrical waves from `singular` to
    `regular`. Not all expansions are available, only those that result in a discrete
    set of modes. For example, plane waves can be expanded in spherical waves, but the
    opposite transformation generally requires a continuous spectrum (an integral) over
    plane waves.

    Args:
        basis (:class:`ScalarBasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        modetype (str, optional): Wave mode, used for
            :class:`ScalarSphericalWaveBasis` (and :class:`ScalarCylindricalWaveBasis`).
        k0 (float, optional): Wave number.
        material (:class:`~acoustotreams.AcousticMaterial` or tuple, optional): Material parameters.
        where (array-like, bool, optional): Only evaluate parts of the expansion matrix,
            the given array must have a shape that matches the output shape.
    """
    if isinstance(basis, (tuple, list)):
        to_basis, basis = basis
    else:
        to_basis = basis
    if isinstance(modetype, (tuple, list)):
        to_modetype, modetype = modetype
    else:
        to_modetype = None
    material = AcousticMaterial(material)
    if isinstance(basis, core.ScalarSphericalWaveBasis) and isinstance(
        to_basis, core.ScalarSphericalWaveBasis
    ):
        modetype = "regular" if modetype is None else modetype
        to_modetype = modetype if to_modetype is None else to_modetype
        return _ssw_ssw_expand(
            basis, to_basis, to_modetype, k0, material, modetype, where
        )
    if isinstance(basis, core.ScalarCylindricalWaveBasis):
        if isinstance(to_basis, core.ScalarCylindricalWaveBasis):
            modetype = "regular" if modetype is None else modetype
            to_modetype = modetype if to_modetype is None else to_modetype
            return _scw_scw_expand(
                basis, to_basis, to_modetype, k0, material, modetype, where
            )
        if isinstance(to_basis, core.ScalarSphericalWaveBasis):
            if modetype != "regular" and to_modetype not in (None, "regular"):
                raise ValueError(f"invalid expansion from {modetype} to {to_modetype}")
            return _ssw_scw_expand(basis, to_basis, k0, material, where)
    if isinstance(basis, core.ScalarPlaneWaveBasis):
        if isinstance(to_basis, core.ScalarPlaneWaveBasis):
            to_modetype = modetype if to_modetype is None else to_modetype
            return _spw_spw_expand(basis, to_basis, k0, material, modetype, where)
        if isinstance(to_basis, core.ScalarCylindricalWaveBasis):
            if to_modetype not in (None, "regular"):
                raise ValueError("invalid modetype")
            return _scw_spw_expand(basis, to_basis, k0, material, modetype, where)
        if isinstance(to_basis, core.ScalarSphericalWaveBasis):
            if to_modetype not in (None, "regular"):
                raise ValueError("invalid modetype")
            return _ssw_spw_expand(
                basis, to_basis, k0, material, modetype, where
            )
    raise TypeError("invalid basis")


class Expand(Operator):
    """Expansion matrix.

    When called as attribute of an object it returns a suitable transformation matrix to
    expand one set of modes into another basis set (and mode type, if applicable).
    See also :func:`expand`.
    """

    _FUNC = staticmethod(expand)

    def __init__(self, basis, modetype=None, *, isinv=False):
        args = (basis,) if modetype is None else (basis, modetype)
        super().__init__(*args, isinv=isinv)

    def __call__(self, **kwargs):
        if self.isinv:
            return self._call_inv(**kwargs)
        args = list(self._args)
        if "basis" in kwargs and not isinstance(args[0], tuple):
            args[0] = (args[0], kwargs.pop("basis"))
        if "modetype" in kwargs and len(args) > 1 and not isinstance(args[1], tuple):
            args[1] = (args[1], kwargs.pop("modetype"))
        return self.FUNC(*args, **kwargs)

    def _call_inv(self, **kwargs):
        args = list(self._args)
        if "basis" in kwargs and not isinstance(args[0], tuple):
            args[0] = (kwargs.pop("basis"), args[0])
        if "modetype" in kwargs and len(args) > 1 and not isinstance(args[1], tuple):
            args[1] = (kwargs.pop("modetype"), args[1])
        return self.FUNC(*args, **kwargs)

    def get_kwargs(self, obj, dim=-1):
        kwargs = super().get_kwargs(obj, dim)
        for name in ("basis", "modetype"):
            val = getattr(obj, name, None)
            if isinstance(val, tuple):
                val = val[dim]
            if val is not None:
                kwargs[name] = val
        return kwargs

    @property
    def inv(self):
        """Inverse expansion.

        The inverse transformation is not available for all transformations.
        """
        if len(self._args) == 1:
            basis, modetype = self._args[0], None
        else:
            basis, modetype = self._args
        if isinstance(basis, tuple):
            basis = tuple(basis[::-1])
        if isinstance(modetype, tuple):
            modetype = tuple(modetype[::-1])
        return type(self)(basis, modetype, isinv=not self.isinv)
    

def _sswl_expand(basis, to_basis, eta, k0, kpar, lattice, material, where):
    """Expand scalar spherical waves in a lattice."""
    ks = k0 * AcousticMaterial().c / material.c
    if lattice.dim == 3:
        try:
            length = len(kpar)
        except TypeError:
            length = 1
        if length == 1:
            lattice = Lattice(lattice, "z")
        elif length == 2:
            lattice = Lattice(lattice, "xy")
        elif length == 3:
            # Last attempt to determine the dimension of the sum
            if np.isnan(kpar[2]):
                lattice = Lattice(lattice, "xy")
            elif np.isnan(kpar[0]) and np.isnan(kpar[1]):
                lattice = Lattice(lattice, "z")
    kpar = WaveVector(kpar)
    if lattice.dim == 1:
        x = kpar[2]
    elif lattice.dim == 2:
        x = kpar[:2]
    else:
        x = kpar[:]
    res = ssw.translate_periodic(
        ks,
        x,
        lattice[...],
        to_basis.positions,
        to_basis[()],
        basis[()],
        basis.positions,
        eta=eta,
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(
        res,
        k0=k0,
        basis=(to_basis, basis),
        material=material,
        lattice=lattice,
        modetype=("regular", "singular"),
        kpar=kpar,
    )


def _scw_ssw_expand(basis, to_basis, k0, kpar, lattice, material, where):
    """Expand scalar spherical waves into scalar cylindrical waves in a lattice."""
    ks = k0 * AcousticMaterial().c / material.c
    where = np.logical_and(where, to_basis.pidx[:, None] == basis.pidx)
    kpar = WaveVector(kpar)
    res = ssw.periodic_to_scw(
        *(m[:, None] for m in to_basis.zm),
        *basis.lm,
        ks,
        Lattice(lattice, "z").volume,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(
        res,
        k0=k0,
        basis=(to_basis, basis),
        material=material,
        modetype="singular",
        lattice=lattice,
        kpar=kpar,
    )


def _spw_ssw_expand(
    basis, to_basis, k0, kpar, lattice, material, modetype, where
):
    """Expand scalar spherical waves into scalar plane waves in a lattice."""
    if modetype is None and isinstance(to_basis, core.ScalarPlaneWaveBasisByComp):
        modetype = "up"
    kpar = WaveVector(kpar)
    kvecs = to_basis.kvecs(k0, material, modetype)
    transl = spw.translate(
        *(b[:, None] for b in kvecs),
        -basis.positions[:, 0],
        -basis.positions[:, 1],
        -basis.positions[:, 2],
    )
    res = transl[:, basis.pidx] * ssw.periodic_to_spw(
        *(b[:, None] for b in kvecs),
        *basis.lm,
        Lattice(lattice, "xy").volume,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(
        res,
        basis=(to_basis, basis),
        k0=k0,
        kpar=kpar,
        lattice=lattice,
        material=material,
        modetype=(modetype, "singular"),
    )


def _scwl_expand(basis, to_basis, eta, k0, kpar, lattice, material, where):
    """Expand scalar cylindrical waves in a lattice."""
    ks = material.ks(k0)
    alignment = (
        "x" if not isinstance(lattice, Lattice) and np.size(lattice) == 1 else None
    )
    lattice = Lattice(lattice, alignment)
    if lattice.dim == 3:
        try:
            length = len(kpar)
        except TypeError:
            length = 1
        if length == 1:
            lattice = Lattice(lattice, "x")
        elif length == 2:
            lattice = Lattice(lattice, "xy")
        elif length == 3:
            # Last attempt to determine the dimension of the sum
            if np.isnan(kpar[1]):
                lattice = Lattice(lattice, "x")
                kpar = WaveVector(kpar, "x")
            else:
                lattice = Lattice(lattice, "xy")
    if lattice.dim == 1:
        kpar = WaveVector(kpar, "x")
        x = kpar[0]
    elif lattice.dim == 2:
        kpar = WaveVector(kpar)
        x = kpar[:2]
    res = scw.translate_periodic(
        ks,
        x,
        lattice[...],
        to_basis.positions,
        to_basis[()],
        basis[()],
        basis.positions,
        eta=eta,
    )
    res[..., np.logical_not(where)] = 0
    for b in (to_basis, basis):
        if b.kpar is not None:
            kpar = kpar & b.kpar
    return core.AcousticsArray(
        res,
        k0=k0,
        basis=(to_basis, basis),
        material=material,
        lattice=lattice,
        modetype=("regular", "singular"),
        kpar=kpar,
    )



def _spw_scw_expand(basis, to_basis, k0, lattice, kpar, material, modetype, where):
    """Expand scalar cylindrical waves into scalar plane waves in a lattice."""
    if modetype is None and isinstance(to_basis, core.ScalarPlaneWaveBasisByComp):
        modetype = "up"
    if len(kpar) == 1:
        kpar = WaveVector(kpar, alignment="x")
    kvecs = to_basis.kvecs(k0, material, modetype)
    transl = spw.translate(
        *(b[:, None] for b in kvecs),
        -basis.positions[:, 0],
        -basis.positions[:, 1],
        -basis.positions[:, 2],
    )
    res = transl[:, basis.pidx] * scw.periodic_to_spw(
        *(b[:, None] for b in kvecs),
        *basis.zm,
        lattice.volume,
        where=where,
    )
    res[..., np.logical_not(where)] = 0
    return core.AcousticsArray(
        res,
        k0=k0,
        basis=(to_basis, basis),
        material=material,
        modetype=(modetype, "singular"),
        lattice=lattice,
        kpar=kpar,
    )

def expandlattice(
    lattice=None,
    kpar=None,
    basis=None,
    modetype=None,
    *,
    eta=0,
    k0=None,
    material=AcousticMaterial(),
    where=True,
):
    """Expansion matrix in lattices.

    Expand the modes from one basis set which are assumed to be periodically repeated on
    a lattice into another basis set.

    Args:
        lattice (:class:`~treams.Lattice` or array-like, optional): Lattice definition.
            In some cases this argument can be omitted, when the lattice can be inferred
            from the basis.
        kpar (sequence, optional): The components of the wave vector tangential to the
            lattice. In some cases this argument can be omitted, when the lattice can be
            inferred from the basis.
        basis (:class:`~acoustotreams.ScalarBasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        k0 (float, optional): Wave number.
        eta (float or complex, optional): Split parameter used when the Ewald summation
            is applied for the lattice sums. By setting it to 0 the split is set
            automatically.
        material (:class:`~acoustotreams.AcousticMaterial` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode, used for
            :class:`ScalarPlaneWaveBasisByComp`.
        where (array-like, bool, optional): Only evaluate parts of the expansion matrix,
            the give array must have a shape that matches the output shape.
    """
    if isinstance(basis, (tuple, list)):
        to_basis, basis = basis
    else:
        to_basis = basis
    if lattice is None and to_basis.lattice is not None:
        lattice = to_basis.lattice
    if not isinstance(lattice, Lattice) and np.size(lattice) == 1:
        alignment = "x" if isinstance(basis, core.ScalarCylindricalWaveBasis) else "z"
    elif isinstance(to_basis, core.ScalarPlaneWaveBasis) and isinstance(
        basis, core.ScalarCylindricalWaveBasis
    ):
        alignment = "x"
    else:
        alignment = None
    lattice = Lattice(lattice, alignment)
    if kpar is None:
        if basis.kpar is None:
            kpar = to_basis.kpar
        else:
            kpar = basis.kpar
    try:
        kpar = list(kpar)
    except TypeError:
        if kpar is None:
            kpar = [np.nan] * 3
        else:
            kpar = [kpar]
    material = AcousticMaterial(material)
    if isinstance(basis, core.ScalarSphericalWaveBasis):
        if isinstance(to_basis, core.ScalarSphericalWaveBasis):
            return _sswl_expand(
                basis, to_basis, eta, k0, kpar, lattice, material, where
            )
        if isinstance(to_basis, core.ScalarCylindricalWaveBasis):
            return _scw_ssw_expand(
                basis, to_basis, k0, kpar, lattice, material, where
            )
        if isinstance(to_basis, core.ScalarPlaneWaveBasis):
            if isinstance(modetype, tuple):
                modetype = modetype[0]
            return _spw_ssw_expand(
                basis, to_basis, k0, kpar, lattice, material, modetype, where
            )
    if isinstance(basis, core.ScalarCylindricalWaveBasis):
        if isinstance(to_basis, core.ScalarCylindricalWaveBasis):
            return _scwl_expand(basis, to_basis, eta, k0, kpar, lattice, material, where)
        if isinstance(to_basis, core.ScalarPlaneWaveBasis):
            if isinstance(modetype, tuple):
                modetype = modetype[0]
            return _spw_scw_expand(
                basis, to_basis, k0, lattice, kpar, material, modetype, where
            )
    raise TypeError("invalid basis")


class ExpandLattice(Operator):
    """Expansion matrix in a lattice.

    When called as attribute of an object it returns a suitable transformation matrix to
    expand one set of modes that is periodically repeated into another basis set.
    See also :func:`expandlattice`.
    """

    _FUNC = staticmethod(expandlattice)

    def __init__(self, lattice=None, kpar=None, basis=None, modetype=None, eta=0):
        super().__init__(lattice, kpar, basis, modetype, eta)

    def __call__(self, **kwargs):
        if self.isinv:
            return self._call_inv(**kwargs)
        args = list(self._args)
        if "basis" in kwargs:
            if args[2] is None:
                args[2] = kwargs.pop("basis")
            else:
                args[2] = (args[2], kwargs.pop("basis"))
        if "modetype" in kwargs and args[3] is None:
            args[3] = kwargs.pop("modetype")
        for i, name in enumerate(("lattice", "kpar")):
            if name in kwargs and args[i] is None:
                args[i] = kwargs.pop(name)
        if args[4] != 0:
            kwargs["eta"] = args[4]
        args = args[:4]
        return self.FUNC(*args, **kwargs)

    def get_kwargs(self, obj, dim=-1):
        kwargs = super().get_kwargs(obj, dim)
        kwargs.pop("modetype", None)
        for name in ("basis", "lattice", "kpar"):
            val = getattr(obj, name, None)
            if isinstance(val, tuple):
                val = val[dim]
            if val is not None:
                kwargs[name] = val
        return kwargs

    @property
    def inv(self):
        """Inverse expansion for periodic arrangements.

        The inverse transformation is not available.
        """
        raise NotImplementedError


def _spwp_permute(basis, n, k0, material, modetype):
    """Permute axes in a scalar partial plane wave basis."""
    if material is None:
        raise TypeError("missing definition of 'material'")
    modetype = "up" if modetype is None else modetype
    kvecs = np.array(basis.kvecs(k0, material, modetype))
    where = (kvecs[..., None] == kvecs[:, None, :]).all(0)
    if n == 1:
        res = spw.permute_xyz(
            kvecs[0],
            kvecs[1],
            kvecs[2],
            where=where,
        )
    elif n == 2:
        res = spw.permute_xyz(
            kvecs[0],
            kvecs[1],
            kvecs[2],
            inverse=True,
            where=where,
        )
    else:
        res = np.eye(len(basis))
    res[np.logical_not(where)] = 0
    return core.AcousticsArray(res, basis=(basis.permute(n), basis))    


def _spwa_permute(basis, n):
    """Permute axes in a scalar plane wave basis."""
    qx, qy, qz = basis.qx, basis.qy, basis.qz
    where = (qx[:, None] == qx) & (qy[:, None] == qy) & (qz[:, None] == qz)
    if n == 1:
        res = spw.permute_xyz(
            qx, qy, qz, where=where
        )
    elif n == 2:
        res = spw.permute_xyz(
            qx,
            qy,
            qz,
            inverse=True,
            where=where,
        )
    elif n == 0:
        res = np.eye(len(basis))
    res[np.logical_not(where)] = 0
    return core.AcousticsArray(res, basis=(basis.permute(n), basis))


def permute(n=1, *, basis, k0=None, material=None, modetype=None):
    """Permutation matrix.

    Permute the axes of a scalar plane wave basis expansion.

    Args:
        n (int, optional): Number of permutations, defaults to 1.
        basis (:class:`~acoustotreams.ScalarBasisSet` or tuple): Basis set, if it is a tuple of two
            basis sets the output and input modes are taken accordingly, else both sets
            of modes are the same.
        k0 (float, optional): Wave number.
        material (:class:`~acoustotreams.AcousticMaterial` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
    """
    if n != int(n):
        raise ValueError("'n' must be integer")
    n = n % 3
    if isinstance(basis, core.ScalarPlaneWaveBasisByComp):
        return _spwp_permute(basis, n, k0, material, modetype)
    if isinstance(basis, core.ScalarPlaneWaveBasisByUnitVector):
        return _spwa_permute(basis, n)
    raise TypeError("invalid basis")


class Permute(Operator):
    """Axes permutation matrix.

    When called as attribute of an object it returns a suitable transformation matrix to
    permute the axis definitions of plane waves. See also :func:`permute`.
    """

    _FUNC = staticmethod(permute)

    def __init__(self, n=1, *, isinv=False):
        super().__init__(n, isinv=isinv)

    @property
    def inv(self):
        return type(self)(*self._args, isinv=not self.isinv)

    def _call_inv(self, **kwargs):
        if "basis" in kwargs:
            kwargs["basis"] = kwargs["basis"].permute(self._args[0])
        return self.FUNC(-self._args[0], **kwargs)
    


def _ssw_vfield(r, basis, k0, material, modetype):
    """Velocity field of scalar spherical waves."""
    ks = k0 * AcousticMaterial().c / material.c
    rsph = sc.car2sph(r - basis.positions)
    res = None
    if modetype == "regular":
        res = wv.vsw_rL(
            basis.l,
            basis.m,
            ks * rsph[..., basis.pidx, 0],
            rsph[..., basis.pidx, 1],
            rsph[..., basis.pidx, 2],
        )
    elif modetype == "singular":
        res = wv.vsw_L(
            basis.l,
            basis.m,
            ks * rsph[..., basis.pidx, 0],
            rsph[..., basis.pidx, 1],
            rsph[..., basis.pidx, 2],
        )
    res *= -1j / (material.rho * material.c)
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(sc.vsph2car(res, rsph[..., basis.pidx, :]))    
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["modetype"] = modetype
    return res


def _scw_vfield(r, basis, k0, material, modetype):
    """Velocity field of scalar cylindrical waves."""
    material = AcousticMaterial(material)
    ks = material.ks(k0)
    krhos = material.krhos(k0, basis.kz)
    krhos[krhos.imag < 0] = -krhos[krhos.imag < 0]
    rcyl = sc.car2cyl(r - basis.positions)
    if modetype == "regular":
        res = wv.vcw_rL(
            basis.kz,
            basis.m,
            krhos * rcyl[..., basis.pidx, 0],
            rcyl[..., basis.pidx, 1],
            rcyl[..., basis.pidx, 2],
            krhos,
            ks
        )
    elif modetype == "singular":
        res = wv.vcw_L(
            basis.kz,
            basis.m,
            krhos * rcyl[..., basis.pidx, 0],
            rcyl[..., basis.pidx, 1],
            rcyl[..., basis.pidx, 2],
            krhos,
            ks
        )   
    res *= -1j / (material.rho * material.c)
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(sc.vcyl2car(res, rcyl[..., basis.pidx, :]))
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["modetype"] = modetype
    return res


def _spw_vfield(r, basis, k0, material, modetype):
    """Velocity field of scalar plane waves."""
    res = None
    kvecs = basis.kvecs(k0, material, modetype)
    res = wv.vpw_L(*kvecs, r[..., 0], r[..., 1], r[..., 2])
    res *= -1j / (material.rho * material.c)
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(res)
    res.ann[-2]["basis"] = basis
    res.ann[-2]["material"] = material
    return res


def vfield(r, *, basis, k0, material=AcousticMaterial(), modetype=None):
    r"""Velocity field.

    The resulting matrix maps the pressure field coefficients of the given basis to the
    velocity field in Cartesian coordinates.

    The velocity field is given in units of :math:`\frac{1}{\rho c}p`.

    Args:
        r (array-like): Evaluation points
        basis (:class:`~acoustotreams.ScalarBasisSet`): Basis set.
        k0 (float): Wave number.
        material (:class:`~acoustotreams.AcousticMaterial` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
    """
    material = AcousticMaterial(material)
    r = np.asanyarray(r)
    r = r[..., None, :]
    if isinstance(basis, core.ScalarSphericalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _ssw_vfield(r, basis, k0, material, modetype).swapaxes(-1, -2)
    if isinstance(basis, core.ScalarCylindricalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _scw_vfield(r, basis, k0, material, modetype).swapaxes(-1, -2)
    if isinstance(basis, core.ScalarPlaneWaveBasis):
        if isinstance(basis, core.ScalarPlaneWaveBasisByComp):
            modetype = "up" if modetype is None else modetype
        return _spw_vfield(r, basis, k0, material, modetype).swapaxes(-1, -2)
    raise TypeError("invalid basis")


class VField(FieldOperator):
    """Velocity field evaluation matrix.

    When called as attribute of an object it returns a suitable matrix to evaluate field
    coefficients at specified points. See also :func:`vfield`.
    """

    _FUNC = staticmethod(vfield)


def _ssw_pfield(r, basis, k0, material, modetype):
    """Pressure field of scalar spherical waves."""
    ks = k0 * AcousticMaterial().c / material.c
    rsph = sc.car2sph(r - basis.positions)
    res = None
    if modetype == "regular":
        res = wv.ssw_rPsi(
            basis.l,
            basis.m,
            ks * rsph[..., basis.pidx, 0],
            rsph[..., basis.pidx, 1],
            rsph[..., basis.pidx, 2],
            )
    elif modetype == "singular":
            res = wv.ssw_Psi(
            basis.l,
            basis.m,
            ks * rsph[..., basis.pidx, 0],
            rsph[..., basis.pidx, 1],
            rsph[..., basis.pidx, 2],
        )
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(np.array([res]).T)  
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["modetype"] = modetype
    return res

def _scw_pfield(r, basis, k0, material, modetype):
    """Pressure field of scalar cylindrical waves."""
    material = AcousticMaterial(material)
    krhos = material.krhos(k0, basis.kz)
    krhos[krhos.imag < 0] = -krhos[krhos.imag < 0]
    rcyl = sc.car2cyl(r - basis.positions)
    if modetype == "regular":
        res = wv.scw_rPsi(
            basis.kz,
            basis.m,
            krhos * rcyl[..., basis.pidx, 0],
            rcyl[..., basis.pidx, 1],
            rcyl[..., basis.pidx, 2],
        )
    elif modetype == "singular":
        res = wv.scw_Psi(
            basis.kz,
            basis.m,
            krhos * rcyl[..., basis.pidx, 0],
            rcyl[..., basis.pidx, 1],
            rcyl[..., basis.pidx, 2]
        )   
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(np.array([res]).T)  
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["modetype"] = modetype
    return res

def _spw_pfield(r, basis, k0, material, modetype):
    """Pressure field of scalar plane waves."""
    res = None
    kvecs = basis.kvecs(k0, material, modetype)
    res = wv.spw_Psi(*kvecs, r[..., 0], r[..., 1], r[..., 2])
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(np.array([res]).T)
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    return res


def pfield(r, *, basis, k0, material=AcousticMaterial(), modetype=None):
    """Pressure field.

    The resulting matrix maps the pressure field coefficients of the given basis to the
    pressure field.

    Args:
        r (array-like): Evaluation points
        basis (:class:`~acoustotreams.ScalarBasisSet`): Basis set.
        k0 (float): Wave number.
        material (:class:`~acoustotreams.AcousticMaterial` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
    """
    material = AcousticMaterial(material)
    r = np.asanyarray(r)
    r = r[..., None, :]
    if isinstance(basis, core.ScalarSphericalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _ssw_pfield(r, basis, k0, material, modetype).swapaxes(-1, -2)
    if isinstance(basis, core.ScalarCylindricalWaveBasis):
        modetype = "regular" if modetype is None else modetype
        return _scw_pfield(r, basis, k0, material, modetype).swapaxes(-1, -2)
    if isinstance(basis, core.ScalarPlaneWaveBasis):
        if isinstance(basis, core.ScalarPlaneWaveBasisByComp):
            modetype = "up" if modetype is None else modetype
        return _spw_pfield(r, basis, k0, material, modetype).swapaxes(-1, -2)
    raise TypeError("invalid basis")


class PField(FieldOperator):
    """Pressure field evaluation matrix.

    When called as attribute of an object it returns a suitable matrix to evaluate field
    coefficients at specified points. See also :func:`pfield`.
    """

    _FUNC = staticmethod(pfield)


def _ssw_pamplitudeff(r, basis, k0, material, modetype):
    """Far-field amplitude of pressure field of singular scalar spherical waves."""
    ks = k0 * AcousticMaterial().c / material.c
    r = sc.car2sph(r)
    rsph = basis.positions
    res = None
    res = wv.ssw_psi(
            basis.l,
            basis.m,
            rsph[..., basis.pidx, 0],
            rsph[..., basis.pidx, 1],
            rsph[..., basis.pidx, 2],
            r[..., basis.pidx, 1],
            r[..., basis.pidx, 2],
            ks
        )
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(np.array([res]).T)  
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["modetype"] = modetype
    return res

def _scw_pamplitudeff(r, basis, k0, material, modetype):
    """Far-field amplitude of pressure field of singular scalar cylindrical waves."""
    r = sc.car2cyl(r)
    material = AcousticMaterial(material)
    krhos = material.krhos(k0, basis.kz)
    krhos[krhos.imag < 0] = -krhos[krhos.imag < 0]
    rcyl = basis.positions
    res = None
    res = wv.scw_psi(
            basis.kz,
            basis.m,
            rcyl[..., basis.pidx, 0],
            rcyl[..., basis.pidx, 1],
            r[..., basis.pidx, 1],
            r[..., basis.pidx, 2],
            krhos,
        )   
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(np.array([res]).T)  
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["modetype"] = modetype
    return res

def pamplitudeff(r, *, basis, k0, material=AcousticMaterial(), modetype=None):
    """Far-field amplitude of pressure field.

    The resulting matrix maps the scattered pressure field coefficients in the scalar 
    spherical or cylindrical wave basis to the far-field amplitude of the pressure field.

    Args:
        r (array-like): Evaluation points
        basis (:class:`~acoustotreams.ScalarBasisSet`): Basis set.
        k0 (float): Wave number.
        material (:class:`~acoustotreams.AcousticMaterial` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
    """
    material = AcousticMaterial(material)
    r = np.asanyarray(r)
    r = r[..., None, :]
    if isinstance(basis, core.ScalarSphericalWaveBasis):
        modetype = "singular" if modetype is None else modetype
        if modetype == "regular":
            raise TypeError("invalid modetype")
        return _ssw_pamplitudeff(r, basis, k0, material, modetype).swapaxes(-1, -2)
    if isinstance(basis, core.ScalarCylindricalWaveBasis):
        modetype = "singular" if modetype is None else modetype
        if modetype == "regular":
            raise TypeError("invalid modetype")
        return _scw_pamplitudeff(r, basis, k0, material, modetype).swapaxes(-1, -2)
    raise TypeError("invalid basis")

class PAmplitudeFF(FieldOperator):
    """Far-field amplitude of pressure field evaluation matrix.

    When called as attribute of an object it returns a suitable matrix to evaluate field
    coefficients at specified points. See also :func:`pamplitudeff`.
    """

    _FUNC = staticmethod(pamplitudeff)


def _ssw_vamplitudeff(r, basis, k0, material, modetype):
    """Far-field amplitude of velocity field of singular vector spherical waves L."""
    ks = k0 * AcousticMaterial().c / material.c
    r = sc.car2sph(r)
    rsph = basis.positions
    res = None
    res = wv.vsw_l(
            basis.l,
            basis.m,
            rsph[..., basis.pidx, 0],
            rsph[..., basis.pidx, 1],
            rsph[..., basis.pidx, 2],
            r[..., basis.pidx, 1],
            r[..., basis.pidx, 2],
            ks
        )
    res *= -1j / (material.rho * material.c)
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(res)    
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["modetype"] = modetype
    return res

def _scw_vamplitudeff(r, basis, k0, material, modetype):
    """Far-field amplitude of pressure field of singular scalar cylindrical waves."""
    ks = k0 * AcousticMaterial().c / material.c
    r = sc.car2cyl(r)
    material = AcousticMaterial(material)
    krhos = material.krhos(k0, basis.kz)
    krhos[krhos.imag < 0] = -krhos[krhos.imag < 0]
    rcyl = basis.positions
    res = None
    res = wv.vcw_l(
            basis.kz,
            basis.m,
            rcyl[..., basis.pidx, 0],
            rcyl[..., basis.pidx, 1],
            r[..., basis.pidx, 1],
            r[..., basis.pidx, 2],
            krhos,
            ks,
        )   
    res *= -1j / (material.rho * material.c)
    if res is None:
        raise ValueError("invalid parameters")
    res = util.AnnotatedArray(res)
    res.ann[-2]["basis"] = basis
    res.ann[-2]["k0"] = k0
    res.ann[-2]["material"] = material
    res.ann[-2]["modetype"] = modetype
    return res

def vamplitudeff(r, *, basis, k0, material=AcousticMaterial(), modetype=None):
    """Far-field amplitude of velocity field.

    The resulting matrix maps the scattered pressure field coefficients in the scalar 
    spherical or cylindrical wave basis to the far-field amplitude of the velocity field
    in spherical or cylindrical coordinates, respectively.

    Args:
        r (array-like): Evaluation points
        basis (:class:`~acoustotreams.ScalarBasisSet`): Basis set.
        k0 (float): Wave number.
        material (:class:`~acoustotreams.AcousticMaterial` or tuple, optional): Material parameters.
        modetype (str, optional): Wave mode.
    """
    material = AcousticMaterial(material)
    r = np.asanyarray(r)
    r = r[..., None, :]
    if isinstance(basis, core.ScalarSphericalWaveBasis):
        modetype = "singular" if modetype is None else modetype
        if modetype == "regular":
            raise TypeError("invalid modetype")
        return _ssw_vamplitudeff(r, basis, k0, material, modetype).swapaxes(-1, -2)
    if isinstance(basis, core.ScalarCylindricalWaveBasis):
        modetype = "singular" if modetype is None else modetype
        if modetype == "regular":
            raise TypeError("invalid modetype")
        return _scw_vamplitudeff(r, basis, k0, material, modetype).swapaxes(-1, -2)
    raise TypeError("invalid basis")

class VAmplitudeFF(FieldOperator):
    """Far-field amplitude of velocity field evaluation matrix.

    When called as attribute of an object it returns a suitable matrix to evaluate field
    coefficients at specified points. See also :func:`vamplitudeff`.
    """

    _FUNC = staticmethod(vamplitudeff)





"""Scattering coefficients for high-symmetry cases.

Calculate the scattering coefficients for cases where they can be obtained analytically
easily. This is a sphere using spherical waves (Mie coefficients), a
cylinder using cylindrical waves, and an infinitely extended planar
interface (Fresnel coefficients).

.. autosummary::
   :toctree:

   mie_acoustics
   mie_acoustics_cyl
   fresnel_acoustics

"""

import treams.special
import numpy as np
from acoustotreams._materialacoustics import AcousticMaterial

def mie_acoustics(l, x, *materials):
    r"""Mie scattering coefficient of degree l.

    The sphere is defined by its size parameter :math:`k_0 r`, where :math:`r` is the
    radius and :math:`k_0` the wave number in air. 

    Likewise, the material parameters are given from inside to outside. These arrays
    are expected to be exactly one unit larger then the array `x`.

    The result is a complex number relating incident with the scattered modes, which are 
    index in the same way.

    Args:
        l (integer): Degree :math:`l \geq 0`
        x (float, array_like): Size parameters
        rho (float or complex, array_like): Mass density
        c (float or complex, array_like): Longitudinal speed of sound
        c_t (float or complex, array_like): Transverse speed of sound

    Returns:
        complex
    """
    mat_sphere, mat_env = zip(*materials) 
    x_env = x * AcousticMaterial().c / mat_env[1]
    j = treams.special.spherical_jn(l, x_env, derivative=False)
    j_d = treams.special.spherical_jn(l, x_env, derivative=True)
    h = treams.special.spherical_hankel1(l, x_env)
    h_d = treams.special.spherical_hankel1_d(l, x_env)
    if np.abs(mat_sphere[2]) == 0 and np.abs(mat_sphere[1]) == 0 and np.abs(mat_sphere[0]) == np.inf:
        res = -j_d / h_d
    elif np.abs(mat_sphere[2]) == 0 and np.abs(mat_sphere[1]) == 0 and np.abs(mat_sphere[0]) == 0:
        res = -j / h
    elif np.abs(mat_sphere[1]) > 0:
        x_sphere = x * AcousticMaterial().c / mat_sphere[1]
        j1 = treams.special.spherical_jn(l, x_sphere, derivative=False)
        j1_d = treams.special.spherical_jn(l, x_sphere, derivative=True)
        if np.abs(mat_sphere[2]) == 0:
            delta = x_sphere * mat_env[0] / (x_env * mat_sphere[0]) 
            res = (delta * j1_d * j - j1 * j_d) / (j1 * h_d - delta * j1_d * h) 
        elif np.abs(mat_sphere[2]) > 0:
            x_sphere_t = x * AcousticMaterial().c / mat_sphere[2]
            j1t = treams.special.spherical_jn(l, x_sphere_t, derivative=False)
            j1t_d = treams.special.spherical_jn(l, x_sphere_t, derivative=True)
            d44 = (l * (l + 1) - 0.5 * x_sphere_t**2) * j1 - 2 * x_sphere * j1_d
            d33 = (l * (l + 1) - 0.5 * x_sphere_t**2 - 1) * j1t - x_sphere_t * j1t_d
            d34 = x_sphere * j1_d - j1
            d43 = l * (l + 1) * (x_sphere_t * j1t_d - j1t)
            d24 = x_sphere * j1_d
            d23 = l * (l + 1) * j1t
            zeta = (d44*d33 - d34*d43) / (d24*d33 - d34*d23)
            res = (-(mat_env[0]/mat_sphere[0] * 0.5 * x_sphere_t**2/x_env * j + zeta * j_d) / 
                (mat_env[0]/mat_sphere[0] * 0.5 * x_sphere_t**2/x_env * h + zeta * h_d))
    return res

def mie_acoustics_cyl(kz, m, k0, radii, *materials):
    r"""Scattering coefficient at an infinite cylinder

    The cylinder is defined by its radii.
    Likewise, the material parameters are given from inside to outside. These arrays
    are expected to be exactly one unit larger then the array `radii`.

    The result is a complex number relating incident with the scattered modes, 
    which are index in the same way.

    Args:
        kz (float): Z component of the wave
        m (integer): Order
        k0 (float or complex): Wave number in vacuum
        radii (float, array_like): Size parameters
        rho (float or complex, array_like): Mass density
        c (float or complex, array_like): Longitudinal speed of sound
        c_t (float or complex, array_like): Transverse speed of sound

    Returns:
        complex
    """

    mat_cyl, mat_env = zip(*materials) 
    k_env = k0 * AcousticMaterial().c / mat_env[1]
    krho_env = np.sqrt(k_env * k_env - kz * kz)
    x_env = krho_env * radii
    j = treams.special.jv(m, x_env)
    j_d = treams.special.jv_d(m, x_env)
    h = treams.special.hankel1(m, x_env)
    h_d = treams.special.hankel1_d(m, x_env)
    
    if np.abs(mat_cyl[2]) == 0 and np.abs(mat_cyl[1]) == 0 and np.abs(mat_cyl[0]) == np.inf:
        res = -j_d / h_d
    elif np.abs(mat_cyl[2]) == 0 and np.abs(mat_cyl[1]) == 0 and np.abs(mat_cyl[0]) == 0:
        res = -j / h
    if np.abs(mat_cyl[1]) > 0:
        k_cyl = k0 * AcousticMaterial().c / mat_cyl[1]
        krho_cyl = np.sqrt(k_cyl * k_cyl - kz * kz)
        x_cyl = krho_cyl * radii
        j1 = treams.special.jv(m, x_cyl)
        j1_d = treams.special.jv_d(m, x_cyl)
        if np.abs(mat_cyl[2]) == 0:
            delta = x_cyl * mat_env[0] / (x_env * mat_cyl[0]) 
            res = (delta * j1_d * j - j1 * j_d) / (j1 * h_d - delta * j1_d * h)
        elif np.abs(mat_cyl[2]) > 0:
            k_cyl_t = k0 * AcousticMaterial().c / mat_cyl[2]
            krho_cyl_t = np.sqrt(k_cyl_t * k_cyl_t - kz * kz)
            x_cyl_t = krho_cyl_t * radii
            j1t = treams.special.jv(m, x_cyl_t)
            j1t_d = treams.special.jv_d(m, x_cyl_t)
            j1t_dd = (-j1t_d + (x_cyl_t**2 + m**2) * j1t) / x_cyl_t
            j1_dd = (-j1_d + (x_cyl**2 + m**2) * j1) / x_cyl
            matrix = np.zeros((3, 3), complex)
            rhs = np.zeros(3, complex)
            lam_0 = mat_env[0] * mat_env[1]**2
            mu_1 = mat_cyl[0] * mat_cyl[2]**2
            lam_1 = mat_cyl[0] * mat_cyl[1]**2 - 2 * mu_1
            matrix[0][1] = 2j * kz * krho_cyl/k_cyl * j1_d
            matrix[0][2] = (krho_cyl_t**2 - kz**2)/k_cyl_t * j1t_d
            matrix[1][0] = k_env * lam_0 * h
            matrix[1][1] = 2 * mu_1 * krho_cyl**2/k_cyl * j1_dd - k_cyl * lam_1 * j1
            matrix[1][2] = 2j * mu_1 * kz * krho_cyl_t/k_cyl_t * j1t_dd
            rhs[1] = -np.power(-1j, m) * k_env * lam_0 * j
            matrix[2][0] = -krho_env/k_env * h_d
            matrix[2][1] = krho_cyl/k_cyl * j1_d
            matrix[2][2] = 1j * kz/k_cyl_t * j1t_d
            rhs[2] = np.power(-1j, m) * krho_env/k_env * j_d
            res = np.linalg.solve(matrix, rhs)
            res = res[0] * np.power(1j, m)
    return res

def fresnel_acoustics(kzs, rhos):
    r"""Fresnel coefficients for a planar interface.

    The first two dimensions index the two media, the second two dimensions
    are added to meet the treams convention.

    The result is an array relating incoming with the outgoing modes, which 
    are indexed in the same way. The first dimension of the array are the outgoing 
    and the second dimension the incoming modes

    .. math::

        \begin{pmatrix}
            t_{\uparrow \uparrow} & r_{\uparrow \downarrow} \\
            r_{\downarrow \uparrow} & t_{\downarrow \downarrow}
        \end{pmatrix}\,

    Args:
        kzs (float): Z component of the waves
        rhos (float or complex): Mass densities

    Returns:
        complex (2, 2, 1, 1)-array
    """,
    res = np.zeros((2, 2, 1, 1), complex)
    res[0][0][0][0] = 2 * rhos[0] * kzs[1] / (rhos[0] * kzs[1] + rhos[1] * kzs[0])
    res[1][1][0][0] = 2 * rhos[1] * kzs[0] / (rhos[0] * kzs[1] + rhos[1] * kzs[0])
    res[0][1][0][0] = (-rhos[0] * kzs[1] + rhos[1] * kzs[0]) / (rhos[0] * kzs[1] + rhos[1] * kzs[0])
    res[1][0][0][0] = -res[0][1]
    return res


"""Scalar spherical wave module.

.. autosummary::
   :toctree:

   periodic_to_scw
   periodic_to_spw
   rotate
   translate
   translate_periodic

"""

import acoustotreams._wavesacoustics as wv
import treams.special as sc
import numpy as np
from treams import lattice
import scipy.special as ss

def _translate_s(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs):
    """Regular translation coefficient for spherical modes"""
    if abs(kr) < 1e-16:
        return 0.0j
    return wv.tl_ssw(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs)

def _translate_r(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs):
    """Singular translation coefficient for spherical modes"""
    return wv.tl_ssw_r(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs)


_translate_s = np.vectorize(_translate_s)
_translate_r = np.vectorize(_translate_r)


def translate(lambda_, mu, l, m, kr, theta, phi, singular=True, *args, **kwargs):
    """Translation coefficient for spherical modes.

    Returns the correct translation coefficient from :func:acoustotreams.tl_ssw,
    and :func:acoustotreams.tl_ssw_r or a combination thereof for the specified mode and
    basis.

    Args:
        lambda_ (int, array_like): Degree of the destination mode
        mu (int, array_like): Order of the destination mode
        l (int, array_like): Degree of the source mode
        m (int, array_like): Order of the source mode
        kr (float or complex, array_like): Translation distance in units of the wave number
        theta (float, array_like): Polar angle
        phi (float, array_like): Azimuthal angle
        singular (bool, optional): If true, singular translation coefficients are used,
            else regular coefficients. Defaults to ``True``.

    Returns:
        complex
    """
    if singular:
        return _translate_s(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs)
    return _translate_r(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs)


def _rotate(lambda_, mu, l, m, phi, theta, psi, *args, **kwargs):
    """
    Rotation coefficient for the rotation around the Euler angles phi, theta and psi.
    It is intended to be used to construct the rotation matrix manually.
    """
    if lambda_ == l:
        return sc.wignerd(l, mu, m, phi, theta, psi, *args, **kwargs)
    else:
        return 0.0j

_rotate = np.vectorize(_rotate)


def rotate(lambda_, mu, l, m, phi, theta=0, psi=0, *args, **kwargs):
    """Rotation coefficient for scalar spherical modes.

    Returns the correct rotation coefficient from :func:`treams.special.wignerd`. The
    angles are given as Euler angles in `z-y-z`-convention. In the intrinsic (object
    fixed coordinate system) convention the rotations are applied in the order phi
    first, theta second, psi third. In the extrinsic (global or reference frame fixed
    coordinate system) the rotations are applied psi first, theta second, phi third.

    Args:
        lambda_ (int, array_like): Degree of the destination mode
        mu (int, array_like): Order of the destination mode
        l (int, array_like): Degree of the source mode
        m (int, array_like): Order of the source mode
        phi (float or complex, array_like): First Euler angle
        theta (float, array_like): Second Euler angle
        psi (float, array_like): Third Euler angle

    Returns:
        complex
    """
    return _rotate(lambda_, mu, l, m, phi, theta, psi, *args, **kwargs)


def _transl_a_lattice(lambda_, mu, l, m, dlms):
    """Singular translation coefficients in a lattice"""
    pref = np.power(-1, np.abs(m)) * np.sqrt(4 * np.pi * (2 * l + 1) * (2 * lambda_ + 1)) * np.power(1.0j, lambda_ - l)
    dlm = 0
    res = 0
    for p in range(l + lambda_, max(abs(lambda_ - l), abs(mu - m)) - 1, -2):
        dlm = dlms[p * (p + 1) + m - mu]
        res += dlm * np.power(1.0j, p) * np.sqrt(2 * p + 1) * sc.wigner3j(l, lambda_, p, m, -mu, -m + mu) * sc.wigner3j(l, lambda_, p, 0, 0, 0)
    return res * pref

def _transl_a_lattice_r(lambda_, mu, l, m, dlms):
    """Regular translation coefficients in a lattice"""
    pref = np.power(-1, np.abs(m)) * np.sqrt(4 * np.pi * (2 * l + 1) * (2 * lambda_ + 1)) * np.power(1.0j, lambda_ - l)
    dlm = 0
    res = 0
    for p in range(l + lambda_, max(abs(lambda_ - l), abs(mu - m)) - 1, -2):
        dlm = dlms[p * (p + 1) + m - mu]
        if p == 0:
            dlm += 1/np.sqrt(4*np.pi)
        res += dlm * np.power(1.0j, p) * np.sqrt(2 * p + 1) * sc.wigner3j(l, lambda_, p, m, -mu, -m + mu) * sc.wigner3j(l, lambda_, p, 0, 0, 0)
    return res * pref

def translate_periodic_r(ks, kpar, a, rs, out, in_=None, rsin=None, eta=0, func=lattice.lsumsw):
    """Regular translation coefficients in a lattice.

    Returns the translation coefficents for the given modes in a lattice. The calculation
    uses the fast converging sums of :py:data:`~treams.lattice`.


    Args:
        ks (float or complex): Wave number of longitudinal waves in the medium
        kpar (float, (D,)-array): Parallel component of the wave, defines the dimension with `1 <= D <= 3`
        a (float, (D,D)-array): Lattice vectors in each row of the array
        rs (float, (M, 3)-array): Shift vectors with respect to one lattice point
        out (2- or 3-tuple of integer arrays): Output modes
        in_ (2- or 3-tuple of integer arrays): Input modes, if none are given equal to
            the output modes
        rsin (float): Shift vectors to use with the input modes, if non are given equal
            to `rs`
        eta (float or complex, optional): Cut between real and reciprocal space
            summation, if equal to zero, an estimation for the optimal value is done.

    Returns:
        complex array
    """

    if in_ is None:
        in_ = out
    out = (*(np.array(o) for o in out),)
    in_ = (*(np.array(i) for i in in_),)
    if len(out) < 2 or len(out) > 3:
        raise ValueError(f"invalid length of output modes {len(out)}, must be 2 or 3")
    elif len(out) == 2:
        out = (np.zeros_like(out[0]),) + out
    if len(in_) < 2 or len(in_) > 3:
        raise ValueError(f"invalid length of input modes {len(in_)}, must be 2 or 3")
    elif len(in_) == 2:
        in_ = (np.zeros_like(in_[0]),) + in_
    if rsin is None:
        rsin = rs
    modes = np.array([
        [l, m]
        for l in range(np.max(out[1]) + np.max(in_[1]) + 1)
        for m in range(-l, l + 1)
    ])
    kpar = np.array(kpar)
    rs = np.array(rs)
    if rs.ndim == 1:
        rs = rs.reshape((1, -1))
    rsin = np.array(rsin)
    if rsin.ndim == 1:
        rsin = rsin.reshape((1, -1))
    rsdiff = -rs[:, None, None, None, :] + rsin[:, None, None, :]

    dim = 1 if kpar.ndim == 0 else kpar.shape[-1]
    # The result has the shape (n_rs, n_rs, n_ks, n_modes)
    dlms = func(dim, modes[:, 0], modes[:, 1], ks, kpar, a, rsdiff, eta)

    res = np.zeros((out[1:][0].shape[-1], in_[1:][0].shape[-1]), np.complex128)
     
    for i in range(out[1:][0].shape[-1]):
        for j in range(in_[1:][0].shape[-1]):
                res[i][j] = _transl_a_lattice_r(in_[1:][0][j], in_[1:][1][j], out[1:][0][i], out[1:][1][i], dlms[0][0][0])
    
    return res

def translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, eta=0, func=lattice.lsumsw):
    """Translation coefficients in a lattice.

    Returns the translation coefficents for the given modes in a lattice. The calculation
    uses the fast converging sums of :py:data:`~treams.lattice`.


    Args:
        ks (float or complex): Wave number of longitudinal waves in the medium
        kpar (float, (D,)-array): Parallel component of the wave, defines the dimension with `1 <= D <= 3`
        a (float, (D,D)-array): Lattice vectors in each row of the array
        rs (float, (M, 3)-array): Shift vectors with respect to one lattice point
        out (2- or 3-tuple of integer arrays): Output modes
        in_ (2- or 3-tuple of integer arrays): Input modes, if none are given equal to
            the output modes
        rsin (float): Shift vectors to use with the input modes, if non are given equal
            to `rs`
        eta (float or complex, optional): Cut between real and reciprocal space
            summation, if equal to zero, an estimation for the optimal value is done.

    Returns:
        complex array
    """

    if in_ is None:
        in_ = out
    out = (*(np.array(o) for o in out),)
    in_ = (*(np.array(i) for i in in_),)
    if len(out) < 2 or len(out) > 3:
        raise ValueError(f"invalid length of output modes {len(out)}, must be 2 or 3")
    elif len(out) == 2:
        out = (np.zeros_like(out[0]),) + out
    if len(in_) < 2 or len(in_) > 3:
        raise ValueError(f"invalid length of input modes {len(in_)}, must be 2 or 3")
    elif len(in_) == 2:
        in_ = (np.zeros_like(in_[0]),) + in_
    if rsin is None:
        rsin = rs
    modes = np.array([
        [l, m]
        for l in range(np.max(out[1]) + np.max(in_[1]) + 1)
        for m in range(-l, l + 1)
    ])
    kpar = np.array(kpar)
    rs = np.array(rs)
    if rs.ndim == 1:
        rs = rs.reshape((1, -1))
    rsin = np.array(rsin)
    if rsin.ndim == 1:
        rsin = rsin.reshape((1, -1))
    rsdiff = -rs[:, None, None, None, :] + rsin[:, None, None, :]

    dim = 1 if kpar.ndim == 0 else kpar.shape[-1]
    # The result has the shape (n_rs, n_rs, n_ks, n_modes)
    dlms = func(dim, modes[:, 0], modes[:, 1], ks, kpar, a, rsdiff, eta)
    res = np.zeros((out[1:][0].shape[-1], in_[1:][0].shape[-1]), np.complex128)
    for i in range(out[1:][0].shape[-1]):
        for j in range(in_[1:][0].shape[-1]):
                res[i][j] = _transl_a_lattice(in_[1:][0][j], in_[1:][1][j], out[1:][0][i], out[1:][1][i], dlms[0][0][0])  
    return res
         

def _periodic_to_spw(kx, ky, kz, l, m, area, *args, **kwargs):
    """Convert periodic spherical wave to plane wave"""
    k = np.sqrt(kx * kx + ky * ky + kz * kz)
    phi = np.arctan2(ky, kx)
    kz_s = kz
    if kz == 0:
        kz_s = 1e-20 + 1e-20j
    elif (np.imag(kz_s) < 0) or (np.imag(kz_s) == 0 and np.real(kz_s) < 0):
        kz_s = -kz_s
    prefactor = (
        np.sqrt(np.pi * (2 * l + 1))
        * np.sqrt(ss.gamma(l - m + 1) / ss.gamma(l + m + 1))
        * np.exp(1j * m * phi)
        * np.power(-1j, l)
        / (area * k * kz_s)
    )
    return prefactor * sc.lpmv(m, l, kz / k + 0j, *args, **kwargs)

_periodic_to_spw = np.vectorize(_periodic_to_spw)


def periodic_to_spw(kx, ky, kz, l, m, area, *args, **kwargs):
    """Convert periodic spherical wave to plane wave.

    Returns the coefficient for the basis change in a periodic arrangement of spherical
    modes to plane waves. For multiple positions only diagonal values (with respect to
    the position) are returned. A correct phase factor is still necessary for the full
    result.

    Args:
        kx (float, array_like): X component of destination mode wave vector
        ky (float, array_like): Y component of destination mode wave vector
        kz (float or complex, array_like): Z component of destination mode wave vector
        l (int, array_like): Degree of the source mode
        m (int, array_like): Order of the source mode
        area (float, array_like): Unit cell area

    Returns:
        complex
    """
    return _periodic_to_spw(kx, ky, kz, l, m, area, *args, **kwargs)

def _periodic_to_scw(kz, mu, l, m, k, a, *args, **kwargs):
    """Convert periodic spherical wave to cylindrical wave"""
    if mu != -m:
        return 0.0j + sc.lpmv(0, 1, 0, *args, **kwargs)
    prefactor = (
        np.sqrt(np.pi * (2 * l + 1), *args, **kwargs) *
        np.sqrt(ss.gamma(l - m + 1) / ss.gamma(l + m + 1), *args, **kwargs) * 
        0.5 * 
        np.power(-1j, l - m, *args, **kwargs)
        / (a * k)
    ) 
    return prefactor * sc.lpmv(m, l, kz/k+0j, *args, **kwargs)

 
_periodic_to_scw = np.vectorize(_periodic_to_scw)


def periodic_to_scw(kz, m, l, mu, k, area, *args, **kwargs):
    """Convert periodic spherical wave to cylindrical wave.

    Returns the coefficient for the basis change in a periodic arrangement of spherical
    modes to cylindrical waves. For multiple positions only diagonal values (with respect to
    the position) are returned. 

    Args:
        kz (float, array_like): Z component of destination mode wave vector
        m (int, array_like): Order or the destination mode
        l (int, array_like): Degree of the source mode
        mu (int, array_like): Order of the source mode
        k (float or complex, array_like): Wave number
        area (float, array_like): Unit cell area

    Returns:
        complex
    """
    return _periodic_to_scw(kz, m, l, mu, k, area, *args, **kwargs)


import treams.special as sc
import numpy as np
import cmath
import scipy.special as ss

def spw_Psi(kx, ky, kz, x, y, z):
    r"""Scalar plane wave :math:`\Psi` 

    The scalar plane wave is defined by

    .. math::

        \psi_{\mathbf k}(\mathbfl r)
        = \mathrm e^{\mathrm i \mathbf k \mathbf r}.

    This function describes a solution to the scalar Helmholtz wave
    equation.

    Args:
        kx (float or complex, array_like): X component of the wave vector
        ky (float or complex, array_like): Y component of the wave vector
        kz (float or complex, array_like): Z component of the wave vector
        x (float, array_like): Y coordinate
        y (float, array_like): X coordinate
        z (float, array_like): Z coordinate

    Returns:
        complex
    
    """
    return np.exp(1j * (kx * x + ky * y + kz * z))


def vpw_L(kx, ky, kz, x, y, z):
    r"""Longitudinal vector plane wave L
    
    The longitudinal vector plane wave is defined by

    .. math::

       \mathbf L_{\mathbf k}(\mathbf r)
        = \frac{\mathbf k}{k} \mathrm e^{\mathrm i \mathbf k \mathbf r}.

    This function describes a longitudinal solution to the vector Helmholtz wave
    equation.

    Args:
        kx (float or complex, array_like): X component of the wave vector
        ky (float or complex, array_like): Y component of the wave vector
        kz (float or complex, array_like): Z component of the wave vector
        x (float, array_like): Y coordinate
        y (float, array_like): X coordinate
        z (float, array_like): Z coordinate

    Returns:
        complex, 3-array
    
    """
    k = np.sqrt(kx * kx + ky * ky + kz * kz)
    phase = np.exp(1j * (kx * x + ky * y + kz * z))
    if k == 0:
        return np.asarray([np.nan, np.nan, np.nan]).T
    else:
        return 1j * np.asarray([kx * phase, ky * phase, kz * phase], np.complex128).T / k

def ssw_Psi(l, m, kr, theta, phi):
    r"""Singular scalar spherical wave :math:`\Psi`

    The singular scalar spherical wave is defined by

    .. math::

        \Psi_{lm}^{(3)}(x, \theta, \varphi)
        = h_l^{(1)}(x) Y_{lm}(\theta, \varphi)

    with :func:`acoustotreams.sph_harm`, and :func:`acoustotreams.spherical_hankel1`.

    This function describes a solution to the scalar Helmholtz wave
    equation.

    Args:
        l (int, array_like): Degree :math:`l \geq 0`
        m (int, array_like): Order :math:`|m| \leq l`
        x (float or complex, array_like): Distance in units of the wave number in the medium :math:`kr`
        theta (float or complex, array_like): Polar angle
        phi (float, array_like): Azimuthal angle

    Returns:
        complex
    """
    return sc.sph_harm(m, l, phi, theta) * sc.spherical_hankel1(l, kr)

def ssw_psi(l, m, x, y, z, theta, phi, k):
     r"""Far-field amplitude of singular scalar spherical wave :math:`\psi`
     Defined by

    .. math::

        \psi_{lm}(x, y, z, \theta, \varphi, k)
        = \frac{\mathrm{i}^{-l-1}}{k} 
        Y_{lm}(\theta, \varphi)
        \mathrm{e}^{-\mathrm{i} k \left(x \sin \theta \cos \varphi + y \sin \theta \sin \varphi + z \cos \theta\right)}

    with :func:`acoustotreams.sph_harm`.

    This function describes a solution to the vector Helmholtz wave
    equation in the limit of :math:`kr \to +\infty`.

    Args:
        l (int, array_like): Degree :math:`l \geq 0`
        m (int, array_like): Order :math:`|m| \leq l`
        x (float or complex, array_like): X coordinate of the source
        y (float or complex, array_like): Y coordinate of the source
        z (float or complex, array_like): Z coordinate of the source
        theta (float or complex, array_like): Polar angle
        phi (float, array_like): Azimuthal angle
        k (float or complex, array_like): Wave number in the medium

    Returns:
        complex
     """
     return (sc.sph_harm(m, l, phi, theta)
             * np.exp(-1j * ((k * x * np.cos(phi) + k * y * np.sin(phi)) * np.sin(theta) 
                             + k * z * np.cos(theta)))
             * np.power(-1j, l + 1)
             * 1 / k)

def ssw_rPsi(l, m, kr, theta, phi):
     r"""Regular scalar spherical wave :math:`\Psi`

     The regular scalar spherical wave is defined by

    .. math::

        \Psi_{lm}^{(1)}(x, \theta, \varphi)
        = j_l(x) Y_{lm}(\theta, \varphi)

    with :func:`acoustotreams.sph_harm`, and :func:`acoustotreams.spherical_jn`.

    This function describes a solution to the scalar Helmholtz wave
    equation.

    Args:
        l (int, array_like): Degree :math:`l \geq 0`
        m (int, array_like): Order :math:`|m| \leq l`
        x (float or complex, array_like): Distance in units of the wave number in the medium :math:`kr`
        theta (float or complex, array_like): Polar angle
        phi (float, array_like): Azimuthal angle

    Returns:
        complex
    """
     return sc.sph_harm(m, l, phi, theta) * sc.spherical_jn(l, kr)


def scw_Psi(kz, m, krr, phi, z):
     r"""Singular scalar cylindrical wave :math:`\Psi`

     The singular scalar cylindrical wave is defined by

     .. math::

         \psi_{k_z m}^{(3)}(x_\rho, \varphi, z)
        = H_m^{(1)}(x_\rho) \mathrm e^{\mathrm i (k_z z + m \varphi)}

     using :func:`acoustotreams.hankel1`.

     This function describes a solution to the scalar Helmholtz wave
     equation.

     Args:
         kz (float, array_like): Z component of the wave vector
         m (int, array_like): Order :math:`|m| \leq l`
         xrho (float or complex, array_like): Radial distance in units of the radial component of the wave number in the medium :math:`k_\rho \rho`
         phi (float, array_like): Azimuthal angle
         z (float, array_like): Z coordinate

     Returns:
         complex
      """
     return np.exp(1j * (m * phi + kz * z)) * sc.hankel1(m, krr)

def scw_psi(kz, m, x, y, phi, z, krho):
     r"""Far-field amplitude of singular scalar cylindrical wave :math:`\psi`

     Defined by

    .. math::

        \psi_{k_z m}(x, y, \varphi, z, k_{\rho})
        = \sqrt{\frac{2}{\pi k_{\rho}}} \mathrm{i}^{-m} \mathrm{e}^{-\mathrm{i} \pi/4}
        \mathrm{e}^{-\mathrm{i} k_{\rho} \left(x \cos \varphi + y \sin \varphi\right)}
        \mathrm{e}^{\mathrm{i} m \varphi + \mathrm{i} k_z z}

    This function describes a solution to the vector Helmholtz wave
    equation in the limit of :math:`k_{\rho} \rho \to +\infty`.

    Args:
        kz (float, array_like): Z component of the wave vector
        m (int, array_like): Order
        x (float or complex, array_like): X coordinate of the source
        y (float or complex, array_like): Y coordinate of the source
        phi (float, array_like): Azimuthal angle
        z (float or complex, array_like): Z coordinate of the source
        krho (float or complex, array_like): Radial component of the wave vector

    Returns:
        complex
     """
     return (np.exp(1j * m * phi + 1j * kz * z) * 
             np.sqrt(2 / (np.pi * krho)) *
             np.exp(-1j * ((krho * x * np.cos(phi) + krho * y * np.sin(phi)))) *
             np.power(-1j, m) * 
             np.exp(-1j * np.pi/4))

def scw_rPsi(kz, m, krr, phi, z):
     r"""Regular scalar cylindrical wave :math:`\Psi`

     The regular scalar cylindrical wave is defined by

     .. math::

         \psi_{k_z m}^{(1)}(x_\rho, \varphi, z)
        = J_m(x_\rho) \mathrm e^{\mathrm i (k_z z + m \varphi)}

     using :func:`acoustotreams.jv`.

     This function describes a solution to the scalar Helmholtz wave
     equation.

     Args:
         kz (float, array_like): Z component of the wave vector
         m (int, array_like): Order :math:`|m| \leq l`
         xrho (float or complex, array_like): Radial distance in units of the radial component of the wave number in the medium :math:`k_\rho \rho`
         phi (float, array_like): Azimuthal angle
         z (float, array_like): Z coordinate

     Returns:
         complex
     """
     return np.exp(1j * (m * phi + kz * z)) * sc.jv(m, krr)


def vsw_L(l, m, kr, theta, phi):
     r"""Singular longitudinal vector spherical wave L

     The singular vector spherical wave is defined by

    .. math::

        \mathbf{L}_{lm}^{(3)}(x, \theta, \varphi)
        = -\mathrm{i} \left[{h_l^{(1)}}'(x) \mathbf{Z}_{lm}(\theta, \varphi) 
        + \frac{h_l^{(1)}(x)}{x} \mathbf{Y}_{lm}(\theta, \varphi) \right]

    with :func:`acoustotreams.vsh_Z`, :func:`acoustotreams.vsh_Y`, 
    :func:`acoustotreams.spherical_hankel1`, and :func:`acoustotreams.spherical_hankel1`.

    This function describes a solution to the vector Helmholtz wave
    equation.

    Args:
        l (int, array_like): Degree :math:`l \geq 0`
        m (int, array_like): Order :math:`|m| \leq l`
        x (float or complex, array_like): Distance in units of the wave number in the medium :math:`kr`
        theta (float or complex, array_like): Polar angle
        phi (float, array_like): Azimuthal angle

    Returns:
        complex, 3-array
     """
     return np.transpose(sc.vsh_Z(l, m, theta, phi).T * sc.spherical_hankel1_d(l, kr)  + np.sqrt(
         l * (l + 1)
     ) * sc.vsh_Y(l, m, theta, phi).T * sc.spherical_hankel1(l, kr) / kr) * (-1.0j)


def vsw_l(l, m, x, y, z, theta, phi, k):
     r"""Far-field amplitude of singular longitudinal vector spherical wave l

     Defined by

    .. math::

        \mathbf{l}_{lm}(x, y, z, \theta, \varphi, k)
        = \frac{\mathrm{i}^{-l}}{k} 
        Y_{lm}(\theta, \varphi)\hat{\mathbf{r}}
        \mathrm{e}^{-\mathrm{i} k \left(x \sin \theta \cos \varphi + y \sin \theta \sin \varphi + z \cos \theta\right)}

    with :func:`acoustotreams.sph_harm`.

    This function describes a solution to the vector Helmholtz wave
    equation in the limit of :math:`kr \to +\infty`.

    Args:
        l (int, array_like): Degree :math:`l \geq 0`
        m (int, array_like): Order :math:`|m| \leq l`
        x (float or complex, array_like): X coordinate of the source
        y (float or complex, array_like): Y coordinate of the source
        z (float or complex, array_like): Z coordinate of the source
        theta (float or complex, array_like): Polar angle
        phi (float, array_like): Azimuthal angle
        k (float or complex, array_like): Wave number in the medium

    Returns:
        complex, 3-array
     """
     return np.transpose(
         np.array(
             [
                 sc.sph_harm(m, l, phi, theta)
                 * np.exp(-1j * ((k * x * np.cos(phi) + k * y * np.sin(phi)) * np.sin(theta) 
                                 + k * z * np.cos(theta)))
                 * np.power(-1j, l)
                 * 1 / k, 
                 np.zeros(len(l)).T,
                 np.zeros(len(l)).T
            ]
        )
    )


def vsw_rL(l, m, kr, theta, phi):
     r"""Regular longitudinal vector spherical wave L

     The regular vector spherical wave is defined by

    .. math::

        \mathbf{L}_{lm}^{(1)}(x, \theta, \varphi)
        = -\mathrm{i} \left[j_l'(x) \mathbf{Z}_{lm}(\theta, \varphi) 
        + \frac{j_l(x)}{x} \mathbf{Y}_{lm}(\theta, \varphi) \right]

    with :func:`acoustotreams.vsh_Z`, :func:`acoustotreams.vsh_Y`, 
    and :func:`acoustotreams.spherical_jn`.

    This function describes a solution to the vector Helmholtz wave
    equation.

    Args:
        l (int, array_like): Degree :math:`l \geq 0`
        m (int, array_like): Order :math:`|m| \leq l`
        x (float or complex, array_like): Distance in units of the wave number in the medium :math:`kr`
        theta (float or complex, array_like): Polar angle
        phi (float, array_like): Azimuthal angle

    Returns:
        complex, 3-array
     """
     res = []
     for i, x in enumerate(kr):
            if x == 0:
                if l[i] != 1:
                    val = np.zeros(3, complex).T
                    res.append(val)
                else:
                   val = np.transpose(
                       sc.vsh_Z(1, m[i], 0, 0).T + 
                       np.sqrt(2) * sc.vsh_Y(1, m[i], 0, 0).T) * (1/3) * (-1.0j)
                   res.append(val)
            else:      
                val = np.transpose(
                    sc.vsh_Z(l[i], m[i], theta[i], phi[i]).T * 
                    sc.spherical_jn_d(l[i], kr[i])  
                    + np.sqrt(l[i] * (l[i] + 1)) * 
                    sc.vsh_Y(l[i], m[i], theta[i], phi[i]).T * 
                    sc.spherical_jn(l[i], kr[i]) / kr[i]) * (-1.0j)
                res.append(val)
     return np.array(res)


def vcw_L(kz, m, krr, phi, z, krho, k):
     r"""Singular longitudinal vector cylindrical wave L

     The singular vector cylindrical wave is defined by

     .. math::

         \mathbf{L}_{k_z m}^{(3)}(x_\rho, \varphi, z, k_{\rho}, k)
        = \left[\frac{k_{\rho}}{k} {H^{(1)}_m(x_\rho)}'(k_{\rho}\rho) \hat{\boldsymbol{\rho}} 
        + \mathrm{i}\frac{m k_{\rho}}{k}\frac{H^{(1)}_m(x_\rho)}{k_{\rho}\rho} \hat{\boldsymbol{\varphi}} 
        + \frac{\mathrm{i} k_z}{k}H^{(1)}_m(x_\rho) \hat{\mathbf{z}} \right] 
        \mathrm{e}^{\mathrm{i} m \varphi + \mathrm{i} k_z z}

     using :func:`acoustotreams.hankel1_d` and :func:`acoustotreams.hankel1`.

     This function describes a longitudinal solution to the vector Helmholtz wave
     equation.

     Args:
         kz (float, array_like): Z component of the wave vector
         m (int, array_like): Order
         xrho (float or complex, array_like): Radial distance in units of the radial component of the wave number in the medium :math:`k_\rho \rho`
         phi (float, array_like): Azimuthal angle
         z (float, array_like): Z coordinate
         krho (float or complex, array_like): Radial component of the wave vector
         k (float or complex, array_like): Wavenumber in the medium 

     Returns:
         complex, 3-array
      """
     return np.transpose(
         np.array(
             [
                 sc.hankel1_d(m, krr) * krho / k  * np.exp(1j * (m * phi + kz * z)),
                 1j * m * sc.hankel1(m, krr) / krr * krho / k  * np.exp(1j * (m * phi + kz * z)),
                 1j * kz / k * sc.hankel1(m, krr) * np.exp(1j * (m * phi + kz * z)),
             ]
         )
     )

def vcw_l(kz, m, x, y, phi, z, krho, k):
     r"""Far-field amplitude of longitudinal singular vector cylindrical wave l

     Defined by

    .. math::

        \mathbf{l}_{k_z m}(x, y, \varphi, z, k_{\rho}, k)
        = \sqrt{\frac{2}{\pi k_{\rho}}} \mathrm{i}^{1-m} \mathrm{e}^{-\mathrm{i} \pi/4}
        \left(\frac{k_{\rho}}{k} \hat{\boldsymbol{\rho}} 
         + \frac{k_{z}}{k} \hat{\mathbf{z}} \right)
        \mathrm{e}^{-\mathrm{i} k_{\rho} \left(x \cos \varphi + y \sin \varphi\right)}
        \mathrm{e}^{\mathrm{i} m \varphi + \mathrm{i} k_z z}

    This function describes a solution to the vector Helmholtz wave
    equation in the limit of :math:`k_{\rho} \rho \to +\infty`.

    Args:
        kz (float, array_like): Z component of the wave vector
        m (int, array_like): Order
        x (float or complex, array_like): X coordinate of the source
        y (float or complex, array_like): Y coordinate of the source
        phi (float, array_like): Azimuthal angle
        z (float or complex, array_like): Z coordinate of the source
        krho (float or complex, array_like): Radial component of the wave vector
        k (float or complex, array_like): Wave number in the medium

    Returns:
        complex, 3-array
     """
     return np.transpose(
         np.array(
             [
                np.sqrt(2 / (np.pi * krho)) *
                np.exp(-1j * ((krho * x * np.cos(phi) + krho * y * np.sin(phi)))) *
                np.power(-1j, m) * 
                np.exp(1j * (m * phi + kz * z)) *
                np.exp(-1j * np.pi/4) *  
                1j * krho / k,
                np.zeros(len(krho)).T,
                np.sqrt(2 / (np.pi * krho)) *
                np.exp(-1j * ((krho * x * np.cos(phi) + krho * y * np.sin(phi)))) *
                np.power(-1j, m) * 
                np.exp(1j * (m * phi + kz * z)) *
                np.exp(-1j * np.pi/4) *  
                1j * kz / k,
             ]
         )
     )

def vcw_rL(kz, m, krr, phi, z, krho, k):
     r"""Regular longitudinal vector cylindrical wave L

     The regular vector cylindrical wave is defined by

     .. math::

         \mathbf{L}_{k_z m}^{(1)}(x_\rho, \varphi, z)
        = \left[\frac{k_{\rho}}{k} {J_m(x_\rho)}'(k_{\rho}\rho) \hat{\boldsymbol{\rho}} 
        + \mathrm{i}\frac{m k_{\rho}}{k}\frac{J_m(x_\rho)}{k_{\rho}\rho} \hat{\boldsymbol{\varphi}} 
        + \frac{\mathrm{i} k_z}{k}J_m(x_\rho) \hat{\mathbf{z}} \right] 
        \mathrm{e}^{\mathrm{i} m \varphi + \mathrm{i} k_z z}

     using :func:`acoustotreams.jv_d` and :func:`acoustotreams.jv`.

     This function describes a longitudinal solution to the vector Helmholtz wave
     equation.

     Args:
         kz (float, array_like): Z component of the wave vector
         m (int, array_like): Order
         xrho (float or complex, array_like): Radial distance in units of the radial component of the wave number in the medium :math:`k_\rho \rho`
         phi (float, array_like): Azimuthal angle
         z (float, array_like): Z coordinate
         krho (float or complex, array_like): Radial component of the wave vector
         k (float or complex, array_like): Wavenumber in the medium 

     Returns:
         complex, 3-array
     """
     res = []
     for i, x in enumerate(krr):
            if x == 0:
                if (abs(m[i]) != 1) and (m[i] != 0):
                    val = np.zeros(3, complex).T
                    res.append(val)
                elif abs(m[i]) == 1:
                   val = np.array(
                                  [
                       0.5 * krho[i] / k[i],
                       0.5 * 1j * m[i] * krho[i] / k[i],
                       0. 
                       ]).T * np.exp(1j * (m[i] * phi[i] + kz[i] * z[i]))
                   res.append(val)
                elif m[i] == 0:
                    val = np.array(
                        [
                            0., 
                            0., 
                            1j * kz[i] / k[i] * np.exp(1j * (m[i] * phi[i] + kz[i] * z[i]))
                            ]
                            ).T
                    res.append(val)
            else:      
                val = np.array(
                [
                    sc.jv_d(m[i], krr[i]) * krho[i] / k[i],
                    1j * m[i] * sc.jv(m[i], krr[i]) / krr[i] * krho[i] / k[i],
                    1j * kz[i] / k[i] * sc.jv(m[i], krr[i]),
                ] 
            ).T * np.exp(1j * (m[i] * phi[i] + kz[i] * z[i]))
                res.append(val)
     return np.array(res)


def _tl_ssw_helper(l, m, lambda_, mu, p, q):
    """Helper function for the translation coefficient of scalar and longitudinal spherical waves"""
    if (
        p < max(abs(m + mu), abs(l - lambda_))
        or p > abs(l + lambda_)
        or q < abs(l - lambda_)
        or q > abs(l + lambda_)
        or (q + l + lambda_) % 2 != 0
    ):
        return 0
    return (
        (2 * p + 1)
        * np.power(1j, lambda_ - l + p)
        * np.sqrt(ss.gamma(p - m - mu + 1) / ss.gamma(p + m + mu + 1))
        * sc.wigner3j(l, lambda_, p, m, mu, -(m + mu))
        * sc.wigner3j(l, lambda_, q, 0, 0, 0)
    )


def tl_ssw(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs):
    r"""Singular translation coefficient of scalar and longitudinal vector spherical waves

        Definded by

    .. math::

        \alpha_{\lambda \mu l m}^{(3)}(x, \theta, \varphi) = (-1)^m \mathrm{i}^{\lambda-l} \sqrt{4\pi (2l+1)(2\lambda+1)} \\
        \cdot \sum_q \mathrm{i}^{q} \sqrt{2q+1} h^{(1)}_q(x) Y_{q,m-\mu}(\theta,\varphi) \\
        \cdot \begin{pmatrix}
            l& \lambda & q \\
            m & -\mu & \mu-m
        \end{pmatrix}
        \begin{pmatrix}
            l& \lambda & q \\
            0& 0& 0
        \end{pmatrix}

    with the Wigner 3j-symbols (:func:`acoustotreams.wigner3j`), the spherical Hankel
    functions (:func:`acoustotreams.spherical_hankel1`), and the spherical harmonics
    (:func:`acoustotreams.sph_harm`). The summation runs over all
    :math:`q \in \{\lambda + l, \lambda + l - 2, \dots, \max(|\lambda - l|, |\mu - m|)\}`.

    These coefficients are used to translate from scattered to incident modes.

    Args:
        lambda (integer, array_like): Degree of the destination mode
        mu (integer, array_like): Order of the destination mode
        l (integer, array_like): Degree of the source mode
        m (integer, array_like): Order of the source mode
        x (complex, array_like): Translation in units of the wave number in the medium
        theta (float or complex, array_like): Polar angle
        phi (float, array_like): Azimuthal angel

    Returns:
        complex
    """
    pref = np.power(-1, np.abs(m)) * np.sqrt((2 * l + 1) * (2 * lambda_ + 1)) * cmath.exp(1j * (m - mu) * phi)
    res = 0.
    max_ = np.max([np.abs(int(lambda_) - int(l)), np.abs(int(m) - int(mu))])
    min_ = int(l) +  int(lambda_)
    for p in range(min_, max_ - 1, -2):
        res += (
            _tl_ssw_helper(l, m, lambda_, -mu, p, p)
            * sc.spherical_hankel1(p, kr)
            * sc.lpmv(m - mu, p, np.cos(theta), *args, **kwargs)
        )
    return res * pref


def tl_ssw_r(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs):
    r"""Regular translation coefficient of scalar and longitudinal vector spherical waves

    Definded by

    .. math::

        \alpha_{\lambda \mu l m}^{(1)}(x, \theta, \varphi) = (-1)^m \mathrm{i}^{\lambda-l} \sqrt{4\pi (2l+1)(2\lambda+1)} \\
        \cdot \sum_q \mathrm{i}^{q} \sqrt{2q+1} j_q(x) Y_{q,m-\mu}(\theta,\varphi) \\
        \cdot \begin{pmatrix}
            l& \lambda & q \\
            m & -\mu & \mu-m
        \end{pmatrix}
        \begin{pmatrix}
            l& \lambda & q \\
            0& 0& 0
        \end{pmatrix}

    with the Wigner 3j-symbols (:func:`acoustotreams.wigner3j`), the spherical Bessel
    functions (:func:`acoustotreams.spherical_jn`), and the spherical harmonics
    (:func:`acoustotreams.sph_harm`). The summation runs over all
    :math:`q \in \{\lambda + l, \lambda + l - 2, \dots, \max(|\lambda - l|, |\mu - m|)\}`.

    These coefficients are used to translate from scattered to scattered modes and
    from incident to incident modes.

    Args:
        lambda (integer, array_like): Degree of the destination mode
        mu (integer, array_like): Order of the destination mode
        l (integer, array_like): Degree of the source mode
        m (integer, array_like): Order of the source mode
        x (complex, array_like): Translation in units of the wave number in the medium
        theta (float or complex, array_like): Polar angle
        phi (float, array_like): Azimuthal angel

    Returns:
        complex
    """
    pref = np.power(-1, np.abs(m)) * np.sqrt((2 * l + 1) * (2 * lambda_ + 1)) * cmath.exp(1j * (m - mu) * phi)
    res = 0.
    max_ = np.max([np.abs(int(lambda_) - int(l)), np.abs(int(m) - int(mu))])
    min_ = int(l) + int(lambda_)
    for p in range(min_, max_ - 1, -2):
        res += (
            _tl_ssw_helper(l, m, lambda_, -mu, p, p)
            * sc.spherical_jn(p, kr)
            * sc.lpmv(m - mu, p, np.cos(theta), *args, **kwargs)
        )
    return res * pref

def tl_scw(kz1, mu, kz2, m, krr, phi, z, *args, **kwargs):
    r"""Singular translation coefficient of scalar and longitudinal vector cylindrical waves

    Definded by

    .. math::

        \begin{cases}
            H_{m - \mu}^{(1)}(x_\rho) \mathrm e^{\mathrm i ((m - \mu) \varphi + k_{z,1}z)}
            & \text{if }k_{z,1} = k_{z,2} \\
            0 & \text{otherwise}
        \end{cases}

    where :math:`H_{m - \mu}^{(1)}` are the Hankel functions (:func:`acoustotreams.hankel1`).

    These coefficients are used to translate from scattered to scattered modes and
    from incident to incident modes.

    Args:
        kz1 (float, array_like): Z component of the destination mode's wave vector
        mu (integer, array_like): Order of the destination mode
        kz2 (float, array_like): Z component of the source mode's wave vector
        m (integer, array_like): Order of the source mode
        xrho (complex, array_like): Translation in radial direction in units of the radial component of the wave number in the medium :math:`k_\rho \rho`
        phi (float, array_like): Azimuthal angel
        z (float, array_like): Translation in z direction

    Returns:
        complex
    """
    if kz1 != kz2:
        return 0.+0.j
    return sc.hankel1(m - mu, krr) * np.exp(1j * ((m - mu) * phi + kz1 * z), *args, **kwargs)


def tl_scw_r(kz1, mu, kz2, m, krr, phi, z, *args, **kwargs):
    r"""Regular translation coefficient of scalar and longitudinal vector cylindrical waves

    Definded by

    .. math::

        \begin{cases}
            J_{m - \mu}(x_\rho) \mathrm e^{\mathrm i ((m - \mu) \varphi + k_{z,1}z)}
            & \text{if }k_{z,1} = k_{z,2} \\
            0 & \text{otherwise}
        \end{cases}

    where :math:`J_{m - \mu}` are the Bessel functions (:func:`acoustotreams.jv`).

    These coefficients are used to translate from scattered to incident modes.

    Args:
        kz1 (float, array_like): Z component of the destination mode's wave vector
        mu (integer, array_like): Order of the destination mode
        kz2 (float, array_like): Z component of the source mode's wave vector
        m (integer, array_like): Order of the source mode
        xrho (complex, array_like): Translation in radial direction in units of the radial component of the wave number in the medium :math:`k_\rho \rho`
        phi (float, array_like): Azimuthal angel
        z (float, array_like): Translation in z direction

    Returns:
        complex
    """
    if kz1 != kz2:
        return 0.+0.j
    return sc.jv(m - mu, krr) * np.exp(1j * ((m - mu) * phi + kz1 * z), *args, **kwargs)